#include "ggap/wswindow-p.h"
#include "ggap/worksheet.h"
#include "ggap/dialogs.h"
#include "ggap/app.h"
#include "ggap/prefs.h"
#include "ui_wswindow.h"
#include <moo/macros.h>
#include <QtGui>

using namespace ggap;

moo::Document *WsWindow::document()
{
    return ws()->mdoc();
}

Worksheet *WsWindow::ws()
{
    return ui->worksheet;
}

void WsWindow::init(bool terminalMode)
{
    ui = new Ui::WsWindow;
    ui->setupUi(this);

    ui->worksheet->setMode(terminalMode ? Worksheet::Terminal : Worksheet::File);

    QObject::connect(ui->worksheet, SIGNAL(copyAvailable(bool)), ui->actionCopy, SLOT(setEnabled(bool)));
    QObject::connect(ui->worksheet, SIGNAL(copyAvailable(bool)), ui->actionCut, SLOT(setEnabled(bool)));
    QObject::connect(ui->worksheet, SIGNAL(copyAvailable(bool)), ui->actionDelete, SLOT(setEnabled(bool)));
    QObject::connect(ui->worksheet, SIGNAL(undoAvailable(bool)), ui->actionUndo, SLOT(setEnabled(bool)));
    QObject::connect(ui->worksheet, SIGNAL(redoAvailable(bool)), ui->actionRedo, SLOT(setEnabled(bool)));

    connectAction(ui->actionUndo, ui->worksheet, SLOT(undo()));
    connectAction(ui->actionRedo, ui->worksheet, SLOT(redo()));
    connectAction(ui->actionCut, ui->worksheet, SLOT(cut()));
    connectAction(ui->actionCopy, ui->worksheet, SLOT(copy()));
    connectAction(ui->actionPaste, ui->worksheet, SLOT(paste()));
    connectAction(ui->actionDelete, ui->worksheet, SLOT(deleteSelected()));
    connectAction(ui->actionDeleteBlock, ui->worksheet, SLOT(deleteBlock()));
    connectAction(ui->actionDeleteAllOutput, ui->worksheet, SLOT(deleteAllOutput()));
    connectAction(ui->actionInsertPromptAfter, ui->worksheet, SLOT(addPromptBlockAfterCursor()));
    connectAction(ui->actionInsertPromptBefore, ui->worksheet, SLOT(addPromptBlockBeforeCursor()));
    connectAction(ui->actionInsertTextAfter, ui->worksheet, SLOT(addTextBlockAfterCursor()));
    connectAction(ui->actionInsertTextBefore, ui->worksheet, SLOT(addTextBlockBeforeCursor()));
    connectAction(ui->actionSelectAll, ui->worksheet, SLOT(selectAll()));
    connectAction(ui->actionInterruptGap, ui->worksheet, SLOT(interruptGap()));
    connectAction(ui->actionRestartGap, ui->worksheet, SLOT(restartGap()));

    connectAction(ui->actionNewWorksheet, SLOT(actionNewWorksheet()));
    connectAction(ui->actionNewTerminal, SLOT(actionNewTerminal()));
    connectAction(ui->actionHelp, gapApp, SLOT(gapHelp()));

    connect(ui->worksheet, SIGNAL(gapStateChanged()), SLOT(gapStateChanged()));

    ui->searchBox->setTextWidget(ui->worksheet);
    connectAction(ui->actionFind, ui->searchBox, SLOT(find()));
    connectAction(ui->actionFindNext, ui->searchBox, SLOT(findNext()));
    connectAction(ui->actionFindPrevious, ui->searchBox, SLOT(findPrevious()));
    ui->searchBox->hide();
    ui->actionReplace->setVisible(false);

    gapStateChanged();

    impl->setupFormatToolbar(ui);
    impl->setupMenuBar(ui);

#ifdef Q_OS_MAC
    ui->actionInterruptGap->setShortcut(QKeySequence(tr("Meta+C")));
#else
    ui->actionInterruptGap->setShortcuts(QList<QKeySequence>() <<
                                            QKeySequence(Qt::CTRL | Qt::Key_Pause) <<
                                            QKeySequence(Qt::CTRL | Qt::Key_Cancel));
#endif

    ui->actionHelp->setShortcuts(QKeySequence::HelpContents);
    ui->actionNewWorksheet->setShortcuts(QKeySequence::New);
    ui->actionUndo->setShortcuts(QKeySequence::Undo);
    ui->actionRedo->setShortcuts(QKeySequence::Redo);
    ui->actionCut->setShortcuts(QKeySequence::Cut);
    ui->actionCopy->setShortcuts(QKeySequence::Copy);
    ui->actionPaste->setShortcuts(QKeySequence::Paste);
    ui->actionHelp->setShortcuts(QKeySequence::HelpContents);
    ui->actionSelectAll->setShortcuts(QKeySequence::SelectAll);
    ui->actionDelete->setShortcuts(QKeySequence::Delete);
    ui->actionPrint->setShortcuts(QKeySequence::Print);
    ui->actionFind->setShortcuts(QKeySequence::Find);
    ui->actionFindNext->setShortcuts(QKeySequence::FindNext);
    ui->actionFindPrevious->setShortcuts(QKeySequence::FindPrevious);
    ui->actionReplace->setShortcuts(QKeySequence::Replace);

    applyPrefs();
    if (terminalMode)
        ui->actionNewTerminal->setVisible(true);

    initUi();
}

void WsWindow::gapStateChanged()
{
    switch (ws()->gapState())
    {
        case GapProcess::Dead:
            ui->statusbar->showMessage(tr("GAP not running"));
            break;
        case GapProcess::Starting:
            ui->statusbar->showMessage(tr("Starting"));
            break;
        case GapProcess::Loading:
            ui->statusbar->showMessage(tr("Loading"));
            break;
        case GapProcess::Busy:
            ui->statusbar->showMessage(tr("Busy"));
            break;
        case GapProcess::InPrompt:
            ui->statusbar->clearMessage();
            break;
    }
}

WsWindow::WsWindow(bool terminalMode) :
    moo::DocumentWindow("worksheet"),
    impl(this), ui(0)
{
    init(terminalMode);
}

WsWindow::~WsWindow()
{
    delete ui;
}

void WsWindow::present(DocumentWindow *old)
{
    DocumentWindow::present(old);

    if (!ws()->isRunning())
        ws()->start();

    // FIXME with the formatting stuff
    ws()->emitCursorPositionChanged();
}


void WsWindow::applyPrefs()
{
    if (ws()->mode() == Worksheet::File)
    {
        bool terminalMode = prefsValue(Prefs::EnableTerminalMode);
        ui->actionNewTerminal->setVisible(terminalMode);
    }

    moo::DocumentWindow::applyPrefs();
}


void WsWindow::actionNewWorksheet()
{
    WsWindow *win = new WsWindow(false);
    win->present(this);
}

void WsWindow::actionNewTerminal()
{
    WsWindow *win = new WsWindow(true);
    win->present(this);
}


QList<QUrl> WsWindow::openFileDialog()
{
    QList<moo::dlg::FileFilterInfo> filters;
    filters << moo::dlg::FileFilterInfo("*.gws *.gwp", "GGAP Files");
    filters << moo::dlg::FileFilterInfo("*.gd *.gi *.g", "GAP Source Files");
    filters << moo::dlg::FileFilterInfo("*", "All files");
    return moo::dlg::openFileNames(this, filters);
}

QUrl WsWindow::saveFileDialog(QString *typeName)
{
    QList<moo::dlg::FileExtensionInfo> extensions;

    if (prefsValue(Prefs::DefaultFileFormat) == "worksheet")
    {
        extensions << moo::dlg::FileExtensionInfo(".gws", "GGAP Worksheet");
        extensions << moo::dlg::FileExtensionInfo(".gwp", "GGAP Workspace");
    }
    else
    {
        extensions << moo::dlg::FileExtensionInfo(".gwp", "GGAP Workspace");
        extensions << moo::dlg::FileExtensionInfo(".gws", "GGAP Worksheet");
    }

    extensions << moo::dlg::FileExtensionInfo(".g", "GAP Source Code");

    moo::dlg::FileExtensionInfo selectedExtension;
    QUrl url = moo::dlg::saveFileName(this, extensions, &selectedExtension);

    if (!url.isEmpty())
    {
        if (selectedExtension.extension == ".gws")
            *typeName = MimeWorksheet;
        else if (selectedExtension.extension == ".gwp")
            *typeName = MimeWorkspace;
        else if (selectedExtension.extension == ".g")
            *typeName = MimePlainText;
        else
        {
            qCritical("%s: oops", Q_FUNC_INFO);
            *typeName = MimeWorkspace;
        }
    }

    return url;
}


void WsWindowPrivate::setupMenuBar(Ui::WsWindow *ui)
{
    M_Q(WsWindow);
    ui->menubar->insertMenu(ui->menuHelp->menuAction(), q->windowMenu());

    ui->menuFile->insertMenu(ui->actionSave, q->recentMenu());

    ui->menuView->addAction(ui->mainToolbar->toggleViewAction());
    QAction *a = formatToolbar->toggleViewAction();
    ui->menuView->addAction(a);
    ui->worksheet->connect(a, SIGNAL(toggled(bool)), SLOT(setTrackCurrentFormat(bool)));
}

static void addActions(QToolBar *toolbar, QActionGroup *group)
{
    QList<QAction*> actions = group->actions();
    foreach (QAction *a, actions)
        toolbar->addAction(a);
}

void WsWindowPrivate::setupFormatToolbar(Ui::WsWindow *ui)
{
    formatToolbar = new QToolBar(pub);
    formatToolbar->hide();
    formatToolbar->setObjectName(QString::fromUtf8("formatToolbar"));
    formatToolbar->setGeometry(QRect(0, 56, 606, 13));
    pub->addToolBar(Qt::TopToolBarArea, formatToolbar);
    pub->insertToolBarBreak(formatToolbar);
    formatToolbar->setWindowTitle(QApplication::translate("WsWindow", "Formatting Toolbar"));

    QToolBar *t = formatToolbar;
    t->addAction(ui->worksheet->actionFontButton(t));
    t->addAction(ui->worksheet->actionBold(t));
    t->addAction(ui->worksheet->actionItalic(t));
    t->addAction(ui->worksheet->actionUnderline(t));
    t->addSeparator();
    addActions(t, ui->worksheet->actionGroupJustify(t));
    t->addSeparator();
    addActions(t, ui->worksheet->actionGroupSupSubScript(t));
    t->addSeparator();
    t->addAction(ui->worksheet->actionTextColor(t));
}

void WsWindow::printDocument(QPrinter *printer)
{
    ws()->print(printer);
}

void WsWindow::setupPrintDialog(QPrintDialog *dlg)
{
    if (ws()->textCursor().hasSelection())
        dlg->addEnabledOption(QAbstractPrintDialog::PrintSelection);
}
